--[[

    TagMe.lua - CHDK script

    The TagMe script allows a user to select a text from a configurable list of text strings
    and write this string into the EXIF UserComment tag (tag ID: 9286.H) of a list of images.

    To define the list of images one has to start the script before the first
    image is taken. TagMe will then save the start image number. Once all images of a list
    has been taken one needs to return (or restart) the script. TagMe will present the numbers
    of the start and end image and one can write a choosen text string to the images EXIF
    UserComment tag.

    TagMe has been developed to simplify the extraction of marked images after download from
    the camera. Typical usage scenarios are marking shots made for time-lapse movies, series of
    HDR or panorama shots. When using SD cards with capacities in the range of several GB
    extracting these series of images out of several thousand can be quite time consuming. With
    a comment text written by TagMe to the EXIF UserComment tag it is possible to use a software
    that can search based on EXIF information (e.g. exiftool) to filter for images with a
    specific comment text.

    PLEASE NOTE: By writing to an EXIF tag TagMe alters the original image file. Please test
    extensively before using the script with your valuable images. This program is distributed
    in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    TagMe has been developed and tested (only) on Canon PowerShot SX10 IS and CHDK firmware
    version 1.03A (sx10-103a-0.9.8-825-full_BETA.zip).


    Author: Stefan Dittforth, (stift at email.de)

    Change History:

        v0.0.1 (15/10/2009) - First version capable to write a selected comment text to
                              the EXIF UserComment tag (9286.H) for a range of images.
        v0.0.2 (04/11/2009) - Amendment. Range of pictures can now spawn multiple directories
                              below A/DCIM/ directory.
                              Added code to restore last access and last modification time stamps
                              after writing UserComment to image.
                              German and english dialog support added.
        v0.0.3 (14/11/2009) - Release version after tests.
        v0.0.4 (14/11/2009) - Changed path for persistence data to "A/CHDK/DATA/TAGME_SAVE.TXT".


CHDK script configuration dialog:

@title Tag Me!
@param l 1 - Deutsch 2 - English
@default l 1

]]

UserCommentTable = {}

-------------------------------------------------------------------------------------
--
-- List of user comments. This list needs to be edited with the keywords that will
-- show up in the comment selection dialog. A comment should not exceed 255 characters
--
-- It appears that the maximum lenght for a user comment on a Canon SX10 IS camera
-- is 255 characters. The function for writing the comment text to the image will
-- trim the text to avoid image file corruption.
--
-------------------------------------------------------------------------------------

UserCommentTable[0] = "Preisverleihung 'Schoenster Gartenzwerg'"
UserCommentTable[1] = "Panorama1"
UserCommentTable[2] = "Panorama2"
UserCommentTable[3] = "Landschaft"
UserCommentTable[4] = "Familie"
UserCommentTable[5] = "Tiere"
UserCommentTable[6] = "HDR1"
UserCommentTable[7] = "HDR2"
UserCommentTable[8] = "HDR3"
UserCommentTable[9] = "Innen"
UserCommentTable[10] = "Aussen"

-------------------------------------------------------------------------------------
--
-- End of user comments. Don't mess around with the code below!
--
-------------------------------------------------------------------------------------

-- add two empty lines before and after, needed to display a "scrollable" list of comments
table.insert(UserCommentTable, 0, "")
table.insert(UserCommentTable, 0, "")
UserCommentTable[table.maxn(UserCommentTable)+1] = " "
UserCommentTable[table.maxn(UserCommentTable)+1] = " "

-- make sure the language variable is within the valid range
if l < 1 then l = 1 end
if l > 2 then l = 2 end

-- The dialog text translation table
-- index 1 - german, index 2 - english
d = {}
d["Start: "] = { [1] = "Start: ", [2] = "Start: " }
d[" Ende: "] = { [1] = " Ende: ", [2] = " End: " }
d["[SET] markieren"] = { [1] = "[SET] markieren", [2] = "[SET] mark images" }
d["[>] waehle Text"] = { [1] = "[>] waehle Text", [2] = "[>] choose text" }
d["[<] reset [v] exit [ALT]"] = { [1] = "[<] reset [v] exit [ALT]", [2] = "[<] reset [v] exit [ALT]" }
d["markiere "] = { [1] = "markiere ", [2] = "marking " }
d["Fertig!"] = { [1] = "Fertig!", [2] = "Done!" }

-------------------------------------------------------------------------------------
--
-- JPEGImage class
--
-- This is the key class of this script. It contains the code to crawl throught the
-- JPEG image file and find the data block with EXIF tag information. To understand
-- this code it is recommended to study the following documents:
--
--       TsuruZoh Tachibanaya, "Description of Exif file format"
--       http://park2.wakwak.com/~tsuruzoh/Computer/Digicams/exif-e.html
--
--       Ger Vermeulen, "Description of Exif file format"
--       http://www.gvsoft.homedns.org/exif/Exif-explanation.html
--
--       JEITA CP-3451, Exchangeable image file format for digital still cameras: Exif Version 2.2
--       http://www.exif.org/Exif2-2.PDF
--
--       TIFF, Revision 6.0, Final  June 3, 1992
--       http://www.exif.org/TIFF6.pdf
--
-------------------------------------------------------------------------------------

JPEGImage = {}

function JPEGImage:new(imageFilePath)

    --
    -- private attributes
    --
    self.FilePath = imageFilePath

    local EXIFDataFormatTable = {}
    EXIFDataFormatTable[1] = { [1] = 1, [2] = "unsigned byte" }
    EXIFDataFormatTable[2] = { [1] = 1, [2] = "ascii strings" }
    EXIFDataFormatTable[3] = { [1] = 2, [2] = "unsigned short" }
    EXIFDataFormatTable[4] = { [1] = 4, [2] = "unsigned long" }
    EXIFDataFormatTable[5] = { [1] = 8, [2] = "unsigned rational" }
    EXIFDataFormatTable[6] = { [1] = 1, [2] = "signed byte" }
    EXIFDataFormatTable[7] = { [1] = 1, [2] = "undefined" }
    EXIFDataFormatTable[8] = { [1] = 2, [2] = "signed short" }
    EXIFDataFormatTable[9] = { [1] = 4, [2] = "signed long" }
    EXIFDataFormatTable[10] = { [1] = 8, [2] = "signed rational" }
    EXIFDataFormatTable[11] = { [1] = 4, [2] = "single float" }
    EXIFDataFormatTable[12] = { [1] = 8, [2] = "double float" }

    --
    -- private methodes
    --

    local function ConvertToBigEndian(InputString)

        local ReturnString = ""
        local InputLength = string.len(InputString)

        for i = 1, InputLength do
            ReturnString = ReturnString..InputString:sub(InputLength-(i-1), InputLength-(i-1))
        end

        return ReturnString

    end

    local function StringToNumber(InputString)

        local ReturnNumber, InputLength, i

        ReturnNumber = 0
        InputLength = string.len(InputString)

        for i = 1, InputLength do
            ReturnNumber = ReturnNumber + InputString:byte(i)*2^((InputLength-i)*8)
        end

        return ReturnNumber

    end

    local function StringToHexCodes(InputString)

        local i, ReturnString

        ReturnString = ""
        for i = 1, string.len(InputString) do
            ReturnString = ReturnString..string.format("%02X", InputString:byte(i))
        end

        return ReturnString

    end

    local function GetBytes(InputString, BigEndian)

        local ReturnString = ""

        if not BigEndian then
            ReturnString = ConvertToBigEndian(InputString)
        else
            ReturnString = InputString
        end

        return ReturnString

    end

    -------------------------------------------------------------------------------------
    --
    -- readIFD (TIFFData, StartIFD)
    --
    -- Input:
    --        TIFFData - TIFF data block, contains all EXIF tags
    --        StartIFD - startposition of IFD in TIFF data block, there can be more than
    --                   one IFD in a TIFF data block
    --
    -- Output: Offset to next IFD
    --
    -- Purpose:
    --         Extracts the EXIF tags from an individual IFD
    --
    -------------------------------------------------------------------------------------

    local function readIFD (TIFFData, StartIFD)

        local NumOfIFDEntries, LenghtIFDEntry, Offset, f, StartNextIFD
        local TagID, DataFormat, NumOfComponents, DataValue, DataStartInFile, FormatDescription

        -- first two bytes of IFD contains number of directory entries
        NumOfIFDEntries = StringToNumber(GetBytes(TIFFData:sub(StartIFD, StartIFD + 1), BigEndian))

        -- read all IFD directory entries (12 bytes per entry)
        LenghtIFDEntry = 12
        Offset = StartIFD + 2

        for i = 0, NumOfIFDEntries - 1 do

            -- get tag number, format and number of components
            TagID               = StringToHexCodes(GetBytes(TIFFData:sub(Offset, Offset + 1), BigEndian))
            DataFormat        = StringToNumber(GetBytes(TIFFData:sub(Offset + 2, Offset + 3), BigEndian))
            FormatDescription = EXIFDataFormatTable[DataFormat][2]
            NumOfComponents   = StringToNumber(GetBytes(TIFFData:sub(Offset + 4, Offset + 7), BigEndian))

            -- get the EXIF data value
            if NumOfComponents * EXIFDataFormatTable[DataFormat][1] > 4 then
                -- next 4 bytes contain an offset to the data
                OffsetToEXIFDataValue = StringToNumber(GetBytes(TIFFData:sub(Offset+8, Offset+11), BigEndian))
                DataValue = TIFFData:sub(1 + OffsetToEXIFDataValue, 1 + OffsetToEXIFDataValue + NumOfComponents)
                DataStartInFile = self.JPEGImage.TIFFHeaderStart + OffsetToEXIFDataValue
            else
                -- next 4 bytes are the data
                DataValue = GetBytes(TIFFData:sub(Offset + 8, Offset + 11), BigEndian)
                DataStartInFile = self.JPEGImage.TIFFHeaderStart + Offset + 8 - 1 -- "- 1" because offset in file starts with 0, in TIFFData it starts with 1
            end

             -- convert EXIF data value
             f = FormatDescription
             if f == "unsigned byte" or f == "unsigned short" or f == "unsigned long" or f == "unsigned rational" then
                 DataValue = StringToNumber(DataValue)
             end

            -- check if we got the EXIF UserComment tag (9286.H)
            if TagID == "9286" then

                -- extract the UserComment string and the start position in the image file
                self.UserComment = DataValue:sub(9, DataValue:len()) -- according to JPEG specificiation first 8 bytes specify code designation
                self.DataStartInFile = DataStartInFile + 8
                self.NumOfComponents = NumOfComponents - 8 -- less then the 8 code designation bytes

                -- we got what we wanted, finish the loop
                break

            end

            -- check if there is an EXIF IFD tag (pointer to EXIF IFD)
            if TagID == "8769" then
                readIFD(TIFFData, DataValue + 1)
            end

             Offset = Offset + LenghtIFDEntry

        end

        -- retun offset to next IFD
        return StringToNumber(GetBytes(TIFFData:sub(Offset, Offset+3), BigEndian))

    end

    -------------------------------------------------------------------------------------
    --
    -- extractEXIFTags (TIFFData)
    --
    -- Input:
    --        TIFFData - TIFF data block, contains all EXIF tags
    --
    -- Output: none
    --
    -- Purpose:
    --         Steps through the IFDs
    --
    -------------------------------------------------------------------------------------

    local function extractEXIFTags(TIFFData)

        local BigEndian, ByteAlignTagMark, StartNextIFD

        -- read byte align from TIFF header
        -- first 2 bytes of TIFF header defines byte align of TIFF data.
        -- 0x4949='I I' is 'Intel' type byte align, 0x4d4d='MM' is 'Motorola' type byte align.
        if StringToHexCodes(TIFFData:sub(1,2)) == "4D4D" then
            BigEndian = true
        elseif StringToHexCodes(TIFFData:sub(1,2)) == "4949" then
            BigEndian = false
        else
            print("byte align setting of TIFF data bytes not found!")
            return nil
        end

        -- test byte align is correct by reading the tag marker
        ByteAlignTagMark = StringToHexCodes(TIFFData:sub(3,4))
        if (BigEndian and ByteAlignTagMark == "002A") or (not BigEndian and ByteAlignTagMark == "2A00") then

            -- the last 4 bytes of TIFF header is the offset to the first IFD
            -- after TIFF header comes first IFD (Image File Directory) aka IFD0
            StartNextIFD = StringToNumber(GetBytes(TIFFData:sub(5,8), BigEndian))

            -- read all IFDs
            repeat
                StartNextIFD = readIFD(TIFFData, StartNextIFD + 1) -- "+ 1" because offset values start from 0 but data block starts from 1
            until StartNextIFD == 0

        else
            print("byte align tag mark not as expected! tag mark = ", ByteAlignTagMark)
            return nil
        end

    end

    -------------------------------------------------------------------------------------
    --
    -- loadEXIFData ()
    --
    -- Checks whether file is a valid JPEG with EXIF data block. Reads TIFF data block
    -- into memory and calls function to extract EXIF tag data.
    --
    -------------------------------------------------------------------------------------

    local function loadEXIFData()

        local inp, data, JPEGMarker, EXIFDataFound, EXIFBlockSize, EXIFData, TIFFData

        inp = io.open(self.FilePath, "rb")
        data = inp:read(2)
        if not data then
            print("could not open file ", self.filePath)
            inp:close()
            return nil
        else
            JPEGMarker = StringToHexCodes(data:sub(1, 2))
            -- the first two bytes of a JPEG file must be "FFD8", the JPEG image marker
            if JPEGMarker == "FFD8" then
                -- search for the EXIF marker "FFE1"
                EXIFDataFound = false
                repeat
                    data = inp:read(1)
                    if data then JPEGMarker = StringToHexCodes(data:sub(1,1)) end
                    if JPEGMarker == "FF" then -- start of a JPEG marker
                        data = inp:read(1)
                        JPEGMarker = StringToHexCodes(data:sub(1,1))
                        if JPEGMarker == "E1" then -- start of EXIF marker found
                            -- initialise the data structure
                            self.JPEGImage = {}
                            -- get the size of the EXIF data block
                            data = inp:read(2)
                            EXIFBlockSize = StringToNumber(data:sub(1, 2))
                            -- check for valid start of EXIF block
                            data = inp:read(6)
                            if StringToHexCodes((data:sub(1, 6))) == "457869660000" then -- valid EXIF block found
                                self.JPEGImage.TIFFHeaderStart = inp:seek() -- save current file position, this is the start of the TIFF header
                                TIFFData = inp:read(EXIFBlockSize - 2 - 6)
                                EXIFDataFound = true
                                -- now that we have loaded all EXIF data close the file
                                inp:close()
                                -- extract EXIF data
                                extractEXIFTags(TIFFData)
                            else
                                print("no valid start of EXIF block found in "..self.filePath.."!")
                                return nil
                            end
                        end
                    end
                until EXIFDataFound or not data
            else
                print("file "..self.filePath.." is not a JPEG image")
                inp:close()
                return nil
            end
        end

    end -- loadEXIFData = function ()

    -- load EXIF data into data structure
    loadEXIFData()

end

function JPEGImage:SetUserComment(NewUserComment)

    local ImageFile, F

    -- Reduce new comment to maximum lenght of UserComment field, if we don't do this we will
    -- corrupt the data structure if the image file.
    -- It looks like on a Canon SX10 IS camera the user comment can be up to 255 characters long.
    if NewUserComment:len() > self.NumOfComponents-1 then NewUserComment = NewUserComment:sub(1, self.NumOfComponents-1) end

    -- Save the time of last access and last modification before writing to the UserComment field
    -- of the image.
    F = os.stat(self.FilePath)

    ImageFile = io.open(self.FilePath, "r+")
    -- move forward to the start position of the user comment field
    ImageFile:seek("set", self.DataStartInFile)
    ImageFile:write(NewUserComment, string.rep(string.char(0), self.NumOfComponents - NewUserComment:len())) -- overwrite the reminder with zeros
    ImageFile:close()

    -- Restore time of last access and last modification.
    -- When writing the new UserComment filed the image file gets a new new "last
    -- mofified" time stamp. With restoring the original time stamps the script
    -- keeps the modifications to an image file to a minimum.
    os.utime(self.FilePath, F["atime"], F["mtime"])

end

-------------------------------------------------------------------------------------
--
-- StoredParameters class
--
-------------------------------------------------------------------------------------

StoredParameters = {}

function StoredParameters:init(FilePath)
    self.StoreFilePath = FilePath
    self.ExpCount = nil
    self.CommentIndex = 0
end

function StoredParameters:read()
    local i
    local f = io.open(self.StoreFilePath,"r")
    if f then
        f:close()
        i = dofile(self.StoreFilePath)
        self.ExpCount = tonumber(i[1])
        self.CommentIndex = tonumber(i[2])
    end
end

function StoredParameters:write()
    local path
    -- If it doesn't exist create the subdirectory where to store persistence data.
    -- Choping off the file name at the end of self.StoreFilePath with pattern match
    -- Gives the folder path. The "- 1" at the end is needed to remove the last "/" from
    -- the path. If we leave "/" in os.mkdir will fail.
    path = string.sub(self.StoreFilePath, 1, string.find(self.StoreFilePath, "/[^/]+$") - 1)
    os.mkdir(path)
    local f = io.open(self.StoreFilePath,"w")
    if f then
        f:write("-- saved data\n")
        f:write("return {"..tostring(self.ExpCount)..", "..tostring(self.CommentIndex).."}")
    else
        print("could not open "..self.StoreFilePath)
        wait_click()
    end
    f:close()
end

-------------------------------------------------------------------------------------
--
-- User dialogs and button control
--
-------------------------------------------------------------------------------------

FrontMenue = {}

function FrontMenue:init()

    self.Line = {} -- list to control which part of the menue are visible and which are hidden
    self.EndPicture = "----"

    self.Line[0] = { [0] = true,  [1] = "" }
    self.Line[1] = { [0] = true,  [1] = "" }
    self.Line[2] = { [0] = false, [1] = "" }
    self.Line[3] = { [0] = true,  [1] = "" }
    self.Line[4] = { [0] = true,  [1] = "" }

    self.ButtonState = {} -- list to control which button are active

    self.ButtonState["UP"] = false
    self.ButtonState["DOWN"] = true
    self.ButtonState["LEFT"] = true
    self.ButtonState["RIGHT"] = true
    self.ButtonState["SET"] = false

end

function FrontMenue:redraw()

    local I, MaxChars, CatText

    -- trim category text to max display lenght
    MaxChars = 25-2 -- space for ">" left and "<" right of the category text needed
    CatText = UserCommentTable[Persistence.CommentIndex]:sub(1, MaxChars)

    -- update the menue with current values
    self.Line[0][1] = d["Start: "][l]..Persistence.ExpCount..d[" Ende: "][l]..self.EndPicture
    self.Line[1][1] = ">"..CatText.."<"
    self.Line[2][1] = d["[SET] markieren"][l]
    self.Line[3][1] = d["[>] waehle Text"][l]
    self.Line[4][1] = d["[<] reset [v] exit [ALT]"][l]

    -- draw the menue
    cls()
    for I = 0, table.maxn(self.Line) do
        if self.Line[I][0] then
            -- the menu line is visible
            print(self.Line[I][1])
        else
            print(" ") -- print an empty line for hidden parts of the menue
        end
    end

end

function SelectUserComment()

    local CurrIndex, i, MaxChars, text, TableLenght, ReturnIndex

    CurrIndex = 0
    MaxLines = 5 -- maximum lines in the CHDK script window
    MaxChars = 25-2 -- space for ">" left and "<" right of the word needed

    repeat

        -- display the list of comments
        cls()
        for i = 0, 4 do
            text = UserCommentTable[CurrIndex+i]:sub(1, MaxChars)
            if i == 2 then
                print(">"..text..string.rep(" ", MaxChars-text:len()).."<")
                ReturnIndex = CurrIndex+i
            else
                print(" "..text..string.rep(" ", MaxChars-text:len()).." ")
            end
        end

        -- wait for user to press button
        wait_click()
        if is_pressed "up" then -- scroll up
            CurrIndex = CurrIndex-1
            if CurrIndex < 0 then CurrIndex = 0 end
        end
        if is_pressed "down" then -- scroll down
            CurrIndex = CurrIndex+1
            TableLenght = table.maxn(UserCommentTable)-(MaxLines-1)
            if CurrIndex >= TableLenght then CurrIndex = TableLenght end
        end

    until is_pressed "set"

    return ReturnIndex

end

function button()
    local key = nil
    wait_click(200)
    if is_pressed "set" then key = "SET" end
    if is_pressed "down" then key = "DOWN" end
    if is_pressed "left" then key = "LEFT" end
    if is_pressed "up" then key = "UP" end
    if is_pressed "right" then key = "RIGHT" end
    return key
end

-------------------------------------------------------------------------------------
--
-- Functions for processing ranges of images
--
-------------------------------------------------------------------------------------

function GetImgDir(FileName)

    -- For a given image file name this function searches for the folder
    -- where the image is stored.

    local F
    local ImgDir = nil
    local Path = "A/DCIM"
    -- get all directory names under A/DCIM
    local Dir = os.listdir(Path, false)
    local Count = #Dir
    local I = 1

    -- remove any directory name that is not "nnnCANON"
    repeat
        local S, E = string.find(Dir[I], "%d%d%dCANON")
        if S == nil and E == nil then table.remove(Dir, I) end
        I = I + 1
    until Dir[I] == nil

    -- sort the table so that the highest folder number is at the end of the table
    table.sort(Dir)

    -- Search for the image, start with the folder with the highest number because
    -- probability to find the image in newer folders is higher.
    I = #Dir
    repeat
        -- test whether file exists
        F = os.stat(Path.."/"..Dir[I].."/"..FileName)
        if not(F == nil) then
            ImgDir = Path.."/"..Dir[I].."/"
        end
        I = I - 1
    until I == 0 or ImgDir ~= nil

    return ImgDir

end

function TagRangeOfImg(S, E, TagText)

    local P = JPEGImage
    local F, I, ImgDir, CurrDir, No, ImgName, ImgFound

    ImgFound = false
    CurrDir = "A/DCIM/"
    for I = S, E, 1 do
        No = tostring(I)
        ImgName = "IMG_"..string.rep("0", 4-string.len(No))..No..".JPG"
        -- test whether image file exists in current image directory
        F = os.stat(CurrDir..ImgName)
        if F == nil then
            -- if file not found search all directories for the image
            ImgDir = GetImgDir(ImgName)
            if ImgDir == nil then
                -- image file not found in any directory
                print(ImgName, " not found!")
                ImgFound = false
            else
                CurrDir = ImgDir
                ImgFound = true
            end
        else
            ImgFound = true
        end
        -- mark image
        if ImgFound then
            print(d["markiere "][l], ImgName)
            P:new(CurrDir..ImgName)
            P:SetUserComment(TagText)
        end
    end

end

-------------------------------------------------------------------------------------
--
-- Main
--
-------------------------------------------------------------------------------------

-- set the file where we can read and save persistence data
Persistence = StoredParameters
Persistence:init("A/CHDK/DATA/TAGME_SAVE.TXT")

Persistence:read() -- get the last saved values
if not Persistence.ExpCount then -- no start image number saved
    Persistence.ExpCount = get_exp_count()+1
    Persistence:write() -- save next exposure count as start image number
end

-- initialise and show the user menue
UserDialog = FrontMenue
UserDialog:init()
UserDialog:redraw()

-- main loop
currentImg2 = nil
repeat

    -- Check whether pictures has been taken since the script started.
    -- This happens when the user pressed ALT to leave the CHDK modus,
    -- takes shots and resturns to the running script by pressing ALT
    -- again.
    currentImg = get_exp_count()
    if currentImg >= Persistence.ExpCount and currentImg ~= currentImg2 then
        UserDialog.EndPicture = currentImg
        -- activate SET menue line if a category has been selected
        if Persistence.CommentIndex > 0 then
            UserDialog.Line[2][0] = true
            UserDialog.ButtonState["SET"] = true
        end
        UserDialog:redraw()
    end

    -- get button input from user and process accordingly
    key = button()

    if key == "RIGHT" and UserDialog.ButtonState[key] then
        -- select category
        Persistence.CommentIndex = SelectUserComment()
        Persistence:write() -- save to SD card
        -- activate SET menue line if there is an end picture number
        if not(UserDialog.EndPicture == "----") then
            UserDialog.Line[2][0] = true
            UserDialog.ButtonState["SET"] = true
        end
        UserDialog:redraw()
    end

    if key == "LEFT" and UserDialog.ButtonState[key] then
        -- reset everything
        Persistence.CommentIndex = 0
        Persistence.ExpCount = get_exp_count()+1
        Persistence:write() -- save reseted values
        -- switch off the SET dialog line and deactivate the SET button
        UserDialog.EndPicture = "----"
        UserDialog.Line[2][0] = false
        UserDialog.ButtonState["SET"] = false
        UserDialog:redraw()
    end

    if key == "SET" and UserDialog.ButtonState[key] then
        -- mark range of pictures with selected tag
        cls()
        TagRangeOfImg(Persistence.ExpCount, UserDialog.EndPicture, UserCommentTable[Persistence.CommentIndex])
        print(d["Fertig!"][l])
        wait_click()
        UserDialog:redraw()
    end

    currentImg2 = currentImg

until key == "DOWN"
